#include <stdio.h>

/* Typedefs */

typedef struct input_source *InputSource;
typedef int ReadProc(void **handle, unsigned char *buf, int max_count);
typedef void CloseProc(void *handle);
typedef int SeekProc(void **handle, int offset);

/* Input sources */

InputSource SourceFromStream(const char8 *description, FILE *file);
InputSource EntityOpen(Entity e);
InputSource NewInputSource(Entity e, ReadProc read, SeekProc seek, 
			   CloseProc close, void *handle);
int SourceTell(InputSource s);
int SourceSeek(InputSource s, int offset);
int SourceLineAndChar(InputSource s, int *linenum, int *charnum);
void SourcePosition(InputSource s, Entity *entity, int *char_number);
int get_with_fill(InputSource s);
void determine_character_encoding(InputSource s);

struct input_source {
    Entity entity;		/* The entity from which the source reads */

    ReadProc *read;
    SeekProc *seek;
    CloseProc *close;
    void *handle;

    Char *line;
    int line_alloc, line_length;
    int next;

    int seen_eoe;
    int complicated_utf8_line;
    int bytes_consumed;
    int bytes_before_current_line;
    int line_end_was_cr;

    int line_number;
    int not_read_yet;

    struct input_source *parent;

    int nextin;
    int insize;
    unsigned char inbuf[4096];
};

#define XEOE (-2)

#define at_eol(s) ((s)->next == (s)->line_length)
#define get(s)    (at_eol(s) ? get_with_fill(s) : (s)->line[(s)->next++])
#define unget(s)  ((s)->seen_eoe ? (s)->seen_eoe= 0 : (s)->next--)
